@@ -10,8 +10,8 @@ import (
 	"strings"
 	"time"
 	"unicode"
-	"unicode/utf8"
 
+	"git.sr.ht/~rockorager/vaxis"
 	"golang.org/x/time/rate"
 )
 
@@ -57,6 +57,7 @@ var SupportedCapabilities = map[string]struct{}{
 	"echo-message":     {},
 	"extended-monitor": {},
 	"invite-notify":    {},
+	"labeled-response": {},
 	"message-tags":     {},
 	"multi-prefix":     {},
 	"server-time":      {},
@@ -66,6 +67,7 @@ var SupportedCapabilities = map[string]struct{}{
 
 	"draft/chathistory":               {},
 	"draft/event-playback":            {},
+	"draft/metadata-2":                {},
 	"draft/read-marker":               {},
 	"soju.im/bouncer-networks-notify": {},
 	"soju.im/bouncer-networks":        {},
@@ -88,14 +90,21 @@ type User struct {
 	Disconnected bool    // can only be true for monitored users.
 }
 
+type ChannelMember struct {
+	Membership string
+	LastActive time.Time
+}
+
 // Channel is a joined channel.
 type Channel struct {
-	Name      string           // the name of the channel.
-	Members   map[*User]string // the set of members associated with their membership.
-	Topic     string           // the topic of the channel, or "" if absent.
-	TopicWho  *Prefix          // the name of the last user who set the topic.
-	TopicTime time.Time        // the last time the topic has been changed.
-	Read      time.Time        // the time until which messages were read.
+	Name      string                  // the name of the channel.
+	Members   map[*User]ChannelMember // the set of members associated with their membership.
+	Topic     string                  // the topic of the channel, or "" if absent.
+	TopicWho  *Prefix                 // the name of the last user who set the topic.
+	TopicTime time.Time               // the last time the topic has been changed.
+	Read      time.Time               // the time until which messages were read.
+	Pinned    bool
+	Muted     bool
 
 	complete bool // whether this structure is fully initialized.
 }
@@ -127,7 +136,10 @@ type Session struct {
 
 	availableCaps map[string]string
 	enabledCaps   map[string]struct{}
+	metadataSubs  map[string]struct{}
 
+	serverName    string
+	defaultPrefix *Prefix
 	// ISUPPORT features
 	casemap       func(string) string
 	chanmodes     [4]string
@@ -138,6 +150,8 @@ type Session struct {
 	prefixModes   string
 	monitor       bool
 	whox          bool
+	listMask      bool
+	upload        string
 
 	users          map[string]*User        // known users.
 	channels       map[string]Channel      // joined channels.
@@ -148,6 +162,7 @@ type Session struct {
 	searchBatchID  string                  // ID of the search targets batch being processed.
 	searchBatch    SearchEvent             // search batch being processed.
 	monitors       map[string]struct{}     // set of users we want to monitor (and keep even if they are disconnected).
+	pendingList    ListEvent               // current list response being received (flushed on list end).
 
 	pendingChannels map[string]time.Time // set of join requests stamps for channels.
 
@@ -168,6 +183,7 @@ func NewSession(out chan<- Message, params SessionParams) *Session {
 		auth:            params.Auth,
 		availableCaps:   map[string]string{},
 		enabledCaps:     map[string]struct{}{},
+		metadataSubs:    map[string]struct{}{},
 		casemap:         CasemapRFC1459,
 		chantypes:       "#&",
 		linelen:         512,
@@ -223,6 +239,34 @@ func (s *Session) HasCapability(capability string) bool {
 	return ok
 }
 
+func (s *Session) IsBouncer() bool {
+	if s.HasCapability("soju.im/bouncer-networks") { // soju-compatible
+		return true
+	}
+	if s.defaultPrefix != nil && s.defaultPrefix.Name == "irc.znc.in" { // ZNC
+		return true
+	}
+	return false
+}
+
+// BouncerService returns the optional nick of the bouncer service user.
+func (s *Session) BouncerService() string {
+	switch s.serverName {
+	case "soju":
+		return "BouncerServ"
+	}
+	return ""
+}
+
+// UploadURL returns the URL to which files can be uploaded according to the FILEHOST specification.
+func (s *Session) UploadURL() string {
+	return s.upload
+}
+
+func (s *Session) HasListMask() bool {
+	return s.listMask
+}
+
 func (s *Session) Nick() string {
 	return s.nick
 }
@@ -265,13 +309,14 @@ func (s *Session) Names(target string) []Member {
 	if s.IsChannel(target) {
 		if c, ok := s.channels[s.Casemap(target)]; ok {
 			names = make([]Member, 0, len(c.Members))
-			for u, pl := range c.Members {
+			for u, m := range c.Members {
 				names = append(names, Member{
-					PowerLevel:   pl,
+					PowerLevel:   m.Membership,
 					Name:         u.Name.Copy(),
 					Away:         u.Away,
 					Disconnected: u.Disconnected,
 					Self:         s.nickCf == s.casemap(u.Name.Name),
+					LastActive:   m.LastActive,
 				})
 			}
 		}
@@ -288,7 +333,10 @@ func (s *Session) Names(target string) []Member {
 			Self: true,
 		})
 	}
-	sort.Sort(members(names))
+	sort.Sort(members{
+		m:        names,
+		prefixes: s.prefixSymbols,
+	})
 	return names
 }
 
@@ -419,20 +467,24 @@ func (s *Session) Away(message string) {
 }
 
 func splitChunks(s string, chunkLen int) (chunks []string) {
-	if chunkLen <= 0 {
+	if chunkLen <= 0 || len(s) <= chunkLen {
 		return []string{s}
 	}
-	for chunkLen < len(s) {
-		i := chunkLen
-		min := chunkLen - utf8.UTFMax
-		for min <= i && !utf8.RuneStart(s[i]) {
-			i--
+
+	b := 0
+	n := 0
+	for _, c := range vaxis.Characters(s) {
+		cw := len(c.Grapheme)
+		if n+cw > chunkLen {
+			chunks = append(chunks, s[b:b+n])
+			b += n
+			n = cw
+			continue
 		}
-		chunks = append(chunks, s[:i])
-		s = s[i:]
+		n += cw
 	}
-	if len(s) != 0 {
-		chunks = append(chunks, s)
+	if b < len(s) {
+		chunks = append(chunks, s[b:])
 	}
 	return
 }
@@ -513,6 +565,34 @@ func (s *Session) ReadSet(target string, timestamp time.Time) {
 	}
 }
 
+func (s *Session) MutedSet(target string, muted bool) (ok bool) {
+	var v string
+	if muted {
+		v = "1"
+	} else {
+		v = "0"
+	}
+	k := "soju.im/muted"
+	if _, ok = s.metadataSubs[k]; ok {
+		s.out <- NewMessage("METADATA", target, "SET", k, v)
+	}
+	return
+}
+
+func (s *Session) PinnedSet(target string, pinned bool) (ok bool) {
+	var v string
+	if pinned {
+		v = "1"
+	} else {
+		v = "0"
+	}
+	k := "soju.im/pinned"
+	if _, ok = s.metadataSubs[k]; ok {
+		s.out <- NewMessage("METADATA", target, "SET", "soju.im/pinned", v)
+	}
+	return
+}
+
 func (s *Session) MonitorAdd(target string) {
 	targetCf := s.casemap(target)
 	if _, ok := s.monitors[targetCf]; !ok {
@@ -631,6 +711,15 @@ func (s *Session) Kick(nick, channel, comment string) {
 }
 
 func (s *Session) HandleMessage(msg Message) (Event, error) {
+	if msg.Prefix == nil {
+		if s.defaultPrefix != nil {
+			msg.Prefix = s.defaultPrefix
+		} else {
+			msg.Prefix = &Prefix{
+				Name: "*",
+			}
+		}
+	}
 	if s.registered {
 		return s.handleRegistered(msg)
 	} else {
@@ -732,6 +821,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			Message:  fmt.Sprintf("Registration failed: %s", strings.Join(msg.Params[1:], " ")),
 		}, nil
 	case rplWelcome:
+		s.defaultPrefix = msg.Prefix
 		if err := msg.ParseParams(&s.nick); err != nil {
 			return nil, err
 		}
@@ -744,6 +834,10 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		if s.host == "" {
 			s.Who(s.nick)
 		}
+	case rplMyinfo:
+		if err := msg.ParseParams(nil, nil, &s.serverName); err != nil {
+			return nil, err
+		}
 	case rplIsupport:
 		if len(msg.Params) < 3 {
 			return nil, msg.errNotEnoughParams(3)
@@ -804,6 +898,16 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 					for channel := range s.channels {
 						s.out <- NewMessage("NAMES", channel)
 					}
+				} else if c.Name == "labeled-response" {
+					if c.Enable {
+						s.out <- Message{
+							Command: labelEnableCommand,
+						}
+					} else {
+						s.out <- Message{
+							Command: labelDisableCommand,
+						}
+					}
 				}
 			}
 		case "NAK":
@@ -823,13 +927,14 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			for _, c := range ParseCaps(caps) {
 				delete(s.availableCaps, c.Name)
 				delete(s.enabledCaps, c.Name)
+				if c.Name == "labeled-response" {
+					s.out <- Message{
+						Command: labelDisableCommand,
+					}
+				}
 			}
 		}
 	case "JOIN":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		var channel string
 		if err := msg.ParseParams(&channel); err != nil {
 			return nil, err
@@ -849,7 +954,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		if s.IsMe(nickCf) {
 			s.channels[channelCf] = Channel{
 				Name:    msg.Params[0],
-				Members: map[*User]string{},
+				Members: map[*User]ChannelMember{},
 			}
 			if _, ok := s.enabledCaps["away-notify"]; ok {
 				// Only try to know who is away if the list is
@@ -861,7 +966,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			if _, ok := s.users[nickCf]; !ok {
 				s.users[nickCf] = &User{Name: msg.Prefix.Copy()}
 			}
-			c.Members[s.users[nickCf]] = ""
+			c.Members[s.users[nickCf]] = ChannelMember{}
 			return UserJoinEvent{
 				User:    msg.Prefix.Name,
 				Channel: c.Name,
@@ -869,10 +974,6 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			}, nil
 		}
 	case "PART":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		var channel string
 		if err := msg.ParseParams(&channel); err != nil {
 			return nil, err
@@ -951,10 +1052,6 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			}
 		}
 	case "QUIT":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		if playback {
 			return UserQuitEvent{
 				User: msg.Prefix.Name,
@@ -1045,7 +1142,9 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 				if _, ok := s.users[nickCf]; !ok {
 					s.users[nickCf] = &User{Name: name.Name.Copy()}
 				}
-				c.Members[s.users[nickCf]] = name.PowerLevel
+				m := c.Members[s.users[nickCf]]
+				m.Membership = name.PowerLevel
+				c.Members[s.users[nickCf]] = m
 			}
 
 			s.channels[channelCf] = c
@@ -1065,6 +1164,8 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 				Channel: c.Name,
 				Topic:   c.Topic,
 				Read:    c.Read,
+				Pinned:  c.Pinned,
+				Muted:   c.Muted,
 			}
 			if stamp, ok := s.pendingChannels[channelCf]; ok && time.Since(stamp) < 5*time.Second {
 				ev.Requested = true
@@ -1112,10 +1213,6 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			s.channels[channelCf] = c
 		}
 	case "TOPIC":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		var channel, topic string
 		if err := msg.ParseParams(&channel, &topic); err != nil {
 			return nil, err
@@ -1126,6 +1223,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 				Channel: channel,
 				Topic:   topic,
 				Time:    msg.TimeOrNow(),
+				Who:     msg.Prefix.Name,
 			}, nil
 		}
 
@@ -1140,6 +1238,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 				Channel: c.Name,
 				Topic:   c.Topic,
 				Time:    msg.TimeOrNow(),
+				Who:     msg.Prefix.Name,
 			}, nil
 		}
 	case "MODE":
@@ -1171,23 +1270,24 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 				}
 				nickCf := s.Casemap(change.Param)
 				user := s.users[nickCf]
-				membership, ok := c.Members[user]
+				m, ok := c.Members[user]
 				if !ok {
 					continue
 				}
 				var newMembership []byte
 				if change.Enable {
-					newMembership = append([]byte(membership), s.prefixSymbols[i])
+					newMembership = append([]byte(m.Membership), s.prefixSymbols[i])
 					sort.Slice(newMembership, func(i, j int) bool {
 						i = strings.IndexByte(s.prefixSymbols, newMembership[i])
 						j = strings.IndexByte(s.prefixSymbols, newMembership[j])
 						return i < j
 					})
-				} else if j := strings.IndexByte(membership, s.prefixSymbols[i]); j >= 0 {
-					newMembership = []byte(membership)
+				} else if j := strings.IndexByte(m.Membership, s.prefixSymbols[i]); j >= 0 {
+					newMembership = []byte(m.Membership)
 					newMembership = append(newMembership[:j], newMembership[j+1:]...)
 				}
-				c.Members[user] = string(newMembership)
+				m.Membership = string(newMembership)
+				c.Members[user] = m
 			}
 			s.channels[channelCf] = c
 			return ModeChangeEvent{
@@ -1197,10 +1297,6 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			}, nil
 		}
 	case "INVITE":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		var nick, channel string
 		if err := msg.ParseParams(&nick, &channel); err != nil {
 			return nil, err
@@ -1223,18 +1319,14 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			Channel: channel,
 		}, nil
 	case "AWAY":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		nickCf := s.Casemap(msg.Prefix.Name)
 
 		if u, ok := s.users[nickCf]; ok {
 			u.Away = len(msg.Params) == 1
 		}
 	case "PRIVMSG", "NOTICE":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
+		if !s.registered && msg.Command == "NOTICE" {
+			return nil, nil
 		}
 
 		var target string
@@ -1242,24 +1334,31 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			return nil, err
 		}
 
-		if playback {
-			return s.newMessageEvent(msg)
-		}
-
 		targetCf := s.casemap(target)
 		nickCf := s.casemap(msg.Prefix.Name)
-		s.typings.Done(targetCf, nickCf)
-
-		return s.newMessageEvent(msg)
+		if !playback {
+			s.typings.Done(targetCf, nickCf)
+		}
+		ev, err := s.newMessageEvent(msg)
+		if err != nil {
+			return nil, err
+		}
+		if c, ok := s.channels[targetCf]; ok {
+			if u, ok := s.users[nickCf]; ok {
+				if m, ok := c.Members[u]; ok {
+					if ev.Time.After(m.LastActive) {
+						m.LastActive = ev.Time
+						c.Members[u] = m
+					}
+				}
+			}
+		}
+		return ev, nil
 	case "TAGMSG":
 		if playback {
 			return nil, nil
 		}
 
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		var target string
 		if err := msg.ParseParams(&target); err != nil {
 			return nil, err
@@ -1329,10 +1428,6 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			}
 		}
 	case "NICK":
-		if msg.Prefix == nil {
-			return nil, errMissingPrefix
-		}
-
 		var nick string
 		if err := msg.ParseParams(&nick); err != nil {
 			return nil, err
@@ -1401,6 +1496,34 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			Target:    target,
 			Timestamp: t,
 		}, nil
+	case "METADATA":
+		// METADATA <Target> <Key> <Visibility> <Value>
+		if len(msg.Params) < 4 {
+			break
+		}
+		var target, key, value string
+		if err := msg.ParseParams(&target, &key, nil, &value); err != nil {
+			return nil, err
+		}
+		channelCf := s.Casemap(target)
+		c, ok := s.channels[channelCf]
+		if !ok {
+			return nil, nil
+		}
+		switch key {
+		case "soju.im/pinned":
+			c.Pinned = value == "1"
+		case "soju.im/muted":
+			c.Muted = value == "1"
+		}
+		s.channels[channelCf] = c
+		if c.complete {
+			return MetadataChangeEvent{
+				Target: target,
+				Pinned: c.Pinned,
+				Muted:  c.Muted,
+			}, nil
+		}
 	case "BOUNCER":
 		if len(msg.Params) < 3 {
 			break
@@ -1435,6 +1558,11 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			return nil, err
 		}
 
+		switch code {
+		case "KEY_INVALID": // METADATA SUB failed: ignore
+			return nil, nil
+		}
+
 		switch msg.Command {
 		case "FAIL":
 			severity = SeverityFail
@@ -1453,7 +1581,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		// silence monlist full error, we don't care because we do it best-effort
 	case rplAway:
 		// we display user away status, we don't care about automatic AWAY replies
-	case rplYourhost, rplCreated, rplMyinfo:
+	case rplYourhost, rplCreated:
 		// useless conection messages
 	case rplAdminme:
 		// useless admin info header
@@ -1465,7 +1593,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		// useless stats delimiter
 	case rplEndofwhois:
 		// useless whois delimiter
-	case rplListstart, rplListend:
+	case rplListstart:
 		// useless list delimiter
 	case rplEndofinvitelist, rplEndofinvexlist:
 		// useless invite list delimiter
@@ -1479,8 +1607,10 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		// useless whois delimiter
 	case rplEndofinfo:
 		// useless info delimiter
-	case rplHelpstart, rplEndofhelp:
+	case rplEndofhelp:
 		// useless help delimiter
+	case rplWhoiskeyvalue, rplKeyvalue, rplKeynotset, rplMetadataunsubok, rplMetadatasubs, rplMetadatasynclater:
+		// useless metadata replies
 	case rplStatscommands:
 		var command, count string
 		if err := msg.ParseParams(nil, &command, &count); err != nil {
@@ -1504,7 +1634,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			Prefix:  "Stats",
 			Message: fmt.Sprintf("The server current uptime is: %s", msg.Params[len(msg.Params)-1]),
 		}, nil
-	case rplLuserclient:
+	case rplStatsconn, rplLuserclient:
 		return InfoEvent{
 			Prefix:  "Stats",
 			Message: msg.Params[len(msg.Params)-1],
@@ -1595,7 +1725,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s %s", nick, text),
+			Message: fmt.Sprintf("%s %s", nick, text),
 		}, nil
 	case rplUnaway:
 		return InfoEvent{
@@ -1612,7 +1742,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s has identified and is registered to the server", nick),
+			Message: fmt.Sprintf("%s has identified and is registered to the server", nick),
 		}, nil
 	case rplWhoisuser:
 		var nick, username, host, realname string
@@ -1621,7 +1751,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s has username %s and host %s (mask %s!%s@%s); their realname is %s", nick, username, host, nick, username, host, realname),
+			Message: fmt.Sprintf("%s has username %s and host %s (mask %s!%s@%s); their realname is %s", nick, username, host, nick, username, host, realname),
 		}, nil
 	case rplWhoisserver:
 		var nick, server, serverInfo string
@@ -1630,16 +1760,16 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s is connected through the server %s (%s)", nick, server, serverInfo),
+			Message: fmt.Sprintf("%s is connected through the server %s (%s)", nick, server, serverInfo),
 		}, nil
 	case rplWhoisoperator:
-		var nick string
-		if err := msg.ParseParams(nil, &nick); err != nil {
+		var nick, opertype string
+		if err := msg.ParseParams(nil, &nick, &opertype); err != nil {
 			return nil, err
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s is an operator of this server", nick),
+			Message: fmt.Sprintf("%s %s", nick, opertype),
 		}, nil
 	case rplWhowasuser:
 		var nick, username, host, realname string
@@ -1648,7 +1778,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s was last seen with username %s and host %s (mask %s!%s@%s); their realname was %s", nick, username, host, nick, username, host, realname),
+			Message: fmt.Sprintf("%s was last seen with username %s and host %s (mask %s!%s@%s); their realname was %s", nick, username, host, nick, username, host, realname),
 		}, nil
 	case rplWhoisidle:
 		var nick, idleText, signonText string
@@ -1665,7 +1795,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		idle := (time.Duration(idleSeconds) * time.Second).String()
 		t := time.Unix(signon, 0)
-		text := fmt.Sprintf("The user %s was idle for %s; they signed-on on %s", nick, idle, t.Local().Format("January 2 at 15:04"))
+		text := fmt.Sprintf("%s was idle for %s; they signed-on on %s", nick, idle, t.Local().Format("January 2 at 15:04"))
 		return InfoEvent{
 			Prefix:  "User",
 			Message: text,
@@ -1677,7 +1807,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s has joined channels: %s", nick, text),
+			Message: fmt.Sprintf("%s has joined channels: %s", nick, text),
 		}, nil
 	case rplWhoisspecial:
 		var nick, text string
@@ -1686,21 +1816,23 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s is also: %s", nick, text),
+			Message: fmt.Sprintf("%s is also: %s", nick, text),
 		}, nil
 	case rplList:
 		var channel, count, topic string
 		if err := msg.ParseParams(nil, &channel, &count, &topic); err != nil {
 			return nil, err
 		}
-		text := fmt.Sprintf("There are %4s users on channel %s", count, channel)
-		if topic != "" {
-			text += " -- " + topic
-		}
-		return InfoEvent{
-			Prefix:  "List",
-			Message: text,
-		}, nil
+		s.pendingList = append(s.pendingList, ListItem{
+			Channel: channel,
+			Count:   count,
+			Topic:   topic,
+		})
+		return nil, nil
+	case rplListend:
+		list := s.pendingList
+		s.pendingList = nil
+		return list, nil
 	case rplChannelmodeis:
 		var channel string
 		if err := msg.ParseParams(nil, &channel); err != nil {
@@ -1732,12 +1864,12 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		if nick != account {
 			return InfoEvent{
 				Prefix:  "User",
-				Message: fmt.Sprintf("The user %s is authenticated as %s", nick, account),
+				Message: fmt.Sprintf("%s is authenticated as %s", nick, account),
 			}, nil
 		} else {
 			return InfoEvent{
 				Prefix:  "User",
-				Message: fmt.Sprintf("The user %s is authenticated", nick),
+				Message: fmt.Sprintf("%s is authenticated", nick),
 			}, nil
 		}
 	case rplInvitelist, rplInvexlist:
@@ -1768,7 +1900,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			}
 			return InfoEvent{
 				Prefix:  "User",
-				Message: fmt.Sprintf("The user %s %s", nick, text),
+				Message: fmt.Sprintf("%s %s", nick, text),
 			}, nil
 		} else if len(msg.Params) >= 4 {
 			var nick string
@@ -1777,7 +1909,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 			}
 			return InfoEvent{
 				Prefix:  "User",
-				Message: fmt.Sprintf("The user %s is actually using the host %s", nick, msg.Params[len(msg.Params)-2]),
+				Message: fmt.Sprintf("%s is actually using the host %s", nick, msg.Params[len(msg.Params)-2]),
 			}, nil
 		}
 	case rplExceptlist:
@@ -1861,7 +1993,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s %s", nick, text),
+			Message: fmt.Sprintf("%s %s", nick, text),
 		}, nil
 	case rplWhoismodes:
 		var nick, text string
@@ -1870,7 +2002,7 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s %s", nick, text),
+			Message: fmt.Sprintf("%s %s", nick, text),
 		}, nil
 	case rplYoureoper:
 		var text string
@@ -1895,9 +2027,17 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 		}
 		return InfoEvent{
 			Prefix:  "User",
-			Message: fmt.Sprintf("The user %s %s", nick, text),
+			Message: fmt.Sprintf("%s %s", nick, text),
 		}, nil
-	case rplHelptxt:
+	case rplMetadatasubok:
+		if err := msg.ParseParams(nil); err != nil {
+			return nil, err
+		}
+		for _, key := range msg.Params[1:] {
+			s.metadataSubs[key] = struct{}{}
+		}
+		return nil, nil
+	case rplHelpstart, rplHelptxt:
 		var text string
 		if err := msg.ParseParams(nil, nil, &text); err != nil {
 			return nil, err
@@ -1917,9 +2057,15 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 				return nil, nil
 
 			}
-			if msg.Command == errUnknowncommand && msg.Params[1] == "BOUNCER" {
-				// ignore any error in response to unconditional BOUNCER LISTNETWORKS
-				return nil, nil
+			if msg.Command == errUnknowncommand {
+				switch msg.Params[1] {
+				case "BOUNCER":
+					// ignore any error in response to unconditional BOUNCER LISTNETWORKS
+					return nil, nil
+				case "METADATA":
+					// ignore any error in response to unconditional METADATA SUB
+					return nil, nil
+				}
 			}
 			return ErrorEvent{
 				Severity: ReplySeverity(msg.Command),
@@ -1932,10 +2078,6 @@ func (s *Session) handleMessageRegistered(msg Message, playback bool) (Event, er
 }
 
 func (s *Session) newMessageEvent(msg Message) (ev MessageEvent, err error) {
-	if msg.Prefix == nil {
-		return ev, errMissingPrefix
-	}
-
 	var target, content string
 	if err := msg.ParseParams(&target, &content); err != nil {
 		return ev, err
@@ -2030,6 +2172,8 @@ func (s *Session) updateFeatures(features []string) {
 			if err == nil {
 				s.historyLimit = historyLimit
 			}
+		case "ELIST":
+			s.listMask = strings.Contains(strings.ToUpper(value), "M")
 		case "LINELEN":
 			linelen, err := strconv.Atoi(value)
 			if err == nil && linelen != 0 {
@@ -2059,6 +2203,8 @@ func (s *Session) updateFeatures(features []string) {
 			s.prefixSymbols = value[numPrefixes+2:]
 		case "WHOX":
 			s.whox = true
+		case "SOJU.IM/FILEHOST":
+			s.upload = value
 		}
 	}
 }
@@ -2067,6 +2213,10 @@ func (s *Session) endRegistration() {
 	if s.registered {
 		return
 	}
+	if len(s.enabledCaps) == 0 || s.HasCapability("draft/metadata-2") {
+		// Best effort to avoid a round trip: subscribe to metadata if explicitly supported or if CAPs are not yet known
+		s.out <- NewMessage("METADATA", "*", "SUB", "soju.im/pinned", "soju.im/muted")
+	}
 	if s.netID != "" {
 		s.out <- NewMessage("BOUNCER", "BIND", s.netID)
 		s.out <- NewMessage("CAP", "END")
